import typing

from apistar import commands, exceptions, hooks, http
from apistar.components import (
    commandline, console, dependency, router, schema, sessions, statics,
    templates, umi
)
from apistar.core import Command, Component
from apistar.frameworks.cli import CliApp
from apistar.interfaces import (
    Auth, CommandLineClient, Console, FileWrapper, Injector, Router, Schema,
    SessionStore, StaticFiles, Templates
)
from apistar.types import Handler, KeywordArgs, UMIChannels, UMIMessage


class ASyncIOApp(CliApp):
    INJECTOR_CLS = dependency.AsyncDependencyInjector

    BUILTIN_COMMANDS = [
        Command('new', commands.new),
        Command('run', commands.run_asyncio),
        Command('schema', commands.schema),
        Command('test', commands.test)
    ]

    BUILTIN_COMPONENTS = [
        Component(Schema, init=schema.CoreAPISchema),
        Component(Templates, init=templates.Jinja2Templates),
        Component(StaticFiles, init=statics.WhiteNoiseStaticFiles),
        Component(Router, init=router.WerkzeugRouter),
        Component(CommandLineClient, init=commandline.ArgParseCommandLineClient),
        Component(Console, init=console.PrintConsole),
        Component(SessionStore, init=sessions.LocalMemorySessionStore),
    ]

    HTTP_COMPONENTS = [
        Component(http.Method, init=umi.get_method),
        Component(http.URL, init=umi.get_url),
        Component(http.Scheme, init=umi.get_scheme),
        Component(http.Host, init=umi.get_host),
        Component(http.Port, init=umi.get_port),
        Component(http.Path, init=umi.get_path),
        Component(http.Headers, init=umi.get_headers),
        Component(http.Header, init=umi.get_header),
        Component(http.QueryString, init=umi.get_querystring),
        Component(http.QueryParams, init=umi.get_queryparams),
        Component(http.QueryParam, init=umi.get_queryparam),
        Component(http.Body, init=umi.get_body),
        Component(http.Request, init=http.Request),
        Component(http.RequestStream, init=umi.get_stream),
        Component(http.RequestData, init=umi.get_request_data),
        Component(FileWrapper, init=umi.get_file_wrapper),
        Component(http.Session, init=sessions.get_session),
        Component(Auth, init=umi.get_auth)
    ]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Setup everything that we need in order to run `self.__call__()`
        self.router = self.preloaded_state[Router]
        self.http_injector = self.create_http_injector()

        settings = kwargs.get('settings', None)

        if settings and 'BEFORE_REQUEST' in settings:
            self.before_request = settings['BEFORE_REQUEST']
        else:
            self.before_request = [hooks.check_permissions_async]

        if settings and 'AFTER_REQUEST' in settings:
            self.after_request = settings['AFTER_REQUEST']
        else:
            self.after_request = [hooks.render_response]

    def create_http_injector(self) -> Injector:
        """
        Create the dependency injector for running handlers in response to
        incoming HTTP requests.
        """
        http_components = {
            component.cls: component.init
            for component in self.HTTP_COMPONENTS
        }

        return self.INJECTOR_CLS(
            components={**http_components, **self.components},
            initial_state=self.preloaded_state,
            required_state={
                UMIMessage: 'message',
                UMIChannels: 'channels',
                KeywordArgs: 'kwargs',
                Handler: 'handler',
                Exception: 'exc',
                http.ResponseHeaders: 'response_headers',
                http.ResponseData: 'response_data'
            },
            resolvers=[dependency.HTTPResolver()]
        )

    async def __call__(self,
                       message: typing.Dict[str, typing.Any],
                       channels: typing.Dict[str, typing.Any]):
        headers = http.ResponseHeaders()
        state = {
            'message': message,
            'channels': channels,
            'handler': None,
            'kwargs': None,
            'exc': None,
            'response_headers': headers
        }
        method = message['method'].upper()
        path = message['path']
        try:
            handler, kwargs = self.router.lookup(path, method)
            state['handler'], state['kwargs'] = handler, kwargs
            funcs = self.before_request + [handler] + self.after_request
            response = await self.http_injector.run_all_async(funcs, state=state)
        except Exception as exc:
            state['exc'] = exc  # type: ignore
            funcs = [self.exception_handler] + self.after_request
            response = await self.http_injector.run_all_async(funcs, state=state)

        headers.update(response.headers)
        if response.content_type is not None:
            headers['content-type'] = response.content_type

        response_message = {
            'status': response.status,
            'headers': [
                [key.encode(), value.encode()]
                for key, value in headers
            ],
            'content': response.content
        }
        await channels['reply'].send(response_message)

    def exception_handler(self, exc: Exception) -> http.Response:
        if isinstance(exc, exceptions.Found):
            return http.Response(
                status=exc.status_code,
                headers={'Location': exc.location}
            )

        if isinstance(exc, exceptions.HTTPException):
            if isinstance(exc.detail, str):
                content = {'message': exc.detail}
            else:
                content = exc.detail
            return http.Response(content, exc.status_code, {})

        raise



import json
import typing

from werkzeug.datastructures import ImmutableMultiDict
from werkzeug.formparser import FormDataParser
from werkzeug.http import parse_options_header
from werkzeug.urls import url_decode

from apistar import exceptions, http


class JSONParser():
    media_type = 'application/json'

    def parse(self, body: http.Body) -> typing.Any:
        if not body:
            raise exceptions.BadRequest(detail='Empty JSON')
        try:
            return json.loads(body.decode('utf-8'))
        except json.JSONDecodeError:
            raise exceptions.BadRequest(detail='Invalid JSON')


class URLEncodedParser():
    media_type = 'application/x-www-form-urlencoded'

    def parse(self, body: http.Body) -> ImmutableMultiDict:
        return url_decode(body, cls=ImmutableMultiDict)


class MultiPartParser():
    media_type = 'multipart/form-data'

    def get_content_length(self, headers: http.Headers) -> typing.Optional[int]:
        content_length = headers.get('Content-Length')
        if content_length is not None:
            try:
                return max(0, int(content_length))
            except (ValueError, TypeError):
                pass
        return None

    def get_mimetype_and_options(self, headers: http.Headers) -> typing.Tuple[str, dict]:
        content_type = headers.get('Content-Type')
        if content_type:
            return parse_options_header(content_type)
        return '', {}

    def parse(self, headers: http.Headers, stream: http.RequestStream) -> ImmutableMultiDict:
        mimetype, options = self.get_mimetype_and_options(headers)
        content_length = self.get_content_length(headers)
        parser = FormDataParser()
        stream, form, files = parser.parse(stream, mimetype, content_length, options)
        return ImmutableMultiDict(list(form.items()) + list(files.items()))


DEFAULT_PARSERS = [
    JSONParser(),
    URLEncodedParser(),
    MultiPartParser()
]



import json
import re
import typing

from werkzeug.http import parse_accept_header

from apistar import http

subtype = re.compile('/[^/]*$')


class Renderer():
    media_type = 'text/plain'
    charset = 'utf-8'  # type: typing.Optional[str]

    def render(self, data: http.ResponseData) -> bytes:
        pass

    def get_content_type(self) -> str:
        if self.charset is None:
            return self.media_type
        return '%s; charset=%s' % (self.media_type, self.charset)


class JSONRenderer(Renderer):
    media_type = 'application/json'
    charset = None

    def render(self, data: http.ResponseData) -> bytes:
        return json.dumps(data).encode('utf-8')


class HTMLRenderer(Renderer):
    media_type = 'text/html'
    charset = 'utf-8'

    def render(self, data: http.ResponseData) -> bytes:
        assert isinstance(data, (str, bytes))
        if isinstance(data, str):
            return data.encode(self.charset)
        return typing.cast(bytes, data)


def negotiate_renderer(accept: typing.Optional[str],
                       renderers: typing.List[Renderer]) -> typing.Optional[Renderer]:
    if accept is None:
        return renderers[0]

    for media_type, quality in parse_accept_header(accept):
        if media_type == '*/*':
            return renderers[0]
        elif media_type.endswith('/*'):
            for renderer in renderers:
                match = re.sub(subtype, '/*', renderer.media_type)
                if media_type == match:
                    return renderer
        else:
            for renderer in renderers:
                if media_type == renderer.media_type:
                    return renderer
    return None


DEFAULT_RENDERERS = [
    JSONRenderer()
]



from apistar.interfaces import Auth


class Unauthenticated(Auth):
    def is_authenticated(self) -> bool:
        return False

    def get_display_name(self) -> None:
        return None

    def get_user_id(self) -> None:
        return None


class Authenticated(Auth):
    def __init__(self, username, user=None, token=None):
        self.username = username
        self.user = user
        self.token = token

    def is_authenticated(self) -> bool:
        return True

    def get_display_name(self) -> str:
        return self.username

    def get_user_id(self) -> str:
        return self.username
